package ch.psi.daq.test.queryrest.controller;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

import javax.annotation.Resource;

import org.junit.After;
import org.junit.Test;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.result.MockMvcResultHandlers;
import org.springframework.test.web.servlet.result.MockMvcResultMatchers;
import org.supercsv.cellprocessor.constraint.NotNull;
import org.supercsv.cellprocessor.ift.CellProcessor;
import org.supercsv.io.CsvMapReader;
import org.supercsv.io.ICsvMapReader;
import org.supercsv.prefs.CsvPreference;

import ch.psi.daq.cassandra.request.range.RequestRangeDate;
import ch.psi.daq.cassandra.request.range.RequestRangePulseId;
import ch.psi.daq.cassandra.request.range.RequestRangeTime;
import ch.psi.daq.cassandra.util.test.CassandraDataGen;
import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.domain.test.TestTimeUtils;
import ch.psi.daq.query.model.Aggregation;
import ch.psi.daq.query.model.AggregationType;
import ch.psi.daq.query.model.Compression;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.query.model.ResponseFormat;
import ch.psi.daq.query.model.impl.DAQQueries;
import ch.psi.daq.query.model.impl.DAQQuery;
import ch.psi.daq.query.model.impl.DAQQueryElement;
import ch.psi.daq.queryrest.controller.QueryRestController;
import ch.psi.daq.queryrest.filter.CorsFilter;
import ch.psi.daq.test.cassandra.admin.CassandraTestAdmin;
import ch.psi.daq.test.queryrest.AbstractDaqRestTest;

/**
 * Tests the {@link DaqController} implementation.
 */
public class QueryRestControllerCsvTest extends AbstractDaqRestTest {

   @Resource
   private CassandraTestAdmin cassandraTestAdmin;

   @Resource
   private CassandraDataGen dataGen;

   @Resource
   private CorsFilter corsFilter;

   public static final String TEST_CHANNEL = "testChannel";
   public static final String TEST_CHANNEL_01 = TEST_CHANNEL + "1";
   public static final String TEST_CHANNEL_02 = TEST_CHANNEL + "2";
   public static final String[] TEST_CHANNEL_NAMES = new String[] {TEST_CHANNEL_01, TEST_CHANNEL_02};

   @After
   public void tearDown() throws Exception {}

   @Test
   public void testPulseRangeQuery() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            TEST_CHANNEL_NAMES);
      request.setResponseFormat(ResponseFormat.CSV);

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      LinkedHashSet<CellProcessor> cellProcessors = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.pulseId);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.shape);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.eventCount);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.value);
      cellProcessors.add(new NotNull());
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      CellProcessor[] processors = cellProcessors.toArray(new CellProcessor[cellProcessors.size()]);

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      ICsvMapReader mapReader =
            new CsvMapReader(new StringReader(response), CsvPreference.EXCEL_NORTH_EUROPE_PREFERENCE);
      try {
         // read comment
         mapReader.read("");
         String[] header = mapReader.getHeader(false);
         Map<String, Object> customerMap;
         long resultsPerChannel = 2;
         long pulse = 0;
         long channelCount = 1;
         while ((customerMap = mapReader.read(header, processors)) != null) {
            assertEquals(TEST_CHANNEL + channelCount, customerMap.get(QueryField.channel.name()));
            assertEquals("" + pulse, customerMap.get(QueryField.pulseId.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocMillis.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalMillis.name()));
            assertEquals("[1]", customerMap.get(QueryField.shape.name()));
            assertEquals("1", customerMap.get(QueryField.eventCount.name()));
            assertEquals("" + pulse, customerMap.get(QueryField.value.name()));

            pulse = ++pulse % resultsPerChannel;
            if (pulse == 0) {
               ++channelCount;
               if (channelCount <= TEST_CHANNEL_NAMES.length) {
                  // read comment (empty rows are skiped
                  mapReader.read("");
                  header = mapReader.getHeader(false);
               }
            }
         }
      } finally {
         mapReader.close();
      }
   }

   @Test
   public void testPulseRangeQueries() throws Exception {
      String testChannel3 = "testChannel3";
      DAQQueries request = new DAQQueries(
            new DAQQueryElement(
                  new RequestRangePulseId(
                        0,
                        1),
                  TEST_CHANNEL_NAMES),
            new DAQQueryElement(
                  new RequestRangePulseId(
                        0,
                        1),
                  testChannel3));
      request.setResponseFormat(ResponseFormat.CSV);

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      LinkedHashSet<CellProcessor> cellProcessors = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.pulseId);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.shape);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.eventCount);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.value);
      cellProcessors.add(new NotNull());
      for (DAQQueryElement element : request) {
         element.setFields(queryFields);
      }

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERIES)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      CellProcessor[] processors = cellProcessors.toArray(new CellProcessor[cellProcessors.size()]);

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      ICsvMapReader mapReader =
            new CsvMapReader(new StringReader(response), CsvPreference.EXCEL_NORTH_EUROPE_PREFERENCE);
      try {
         // read comment
         mapReader.read("");
         String[] header = mapReader.getHeader(false);
         Map<String, Object> customerMap;
         long resultsPerChannel = 2;
         long pulse = 0;
         long channelCount = 1;
         while ((customerMap = mapReader.read(header, processors)) != null) {
            assertEquals(TEST_CHANNEL + channelCount, customerMap.get(QueryField.channel.name()));
            assertEquals("" + pulse, customerMap.get(QueryField.pulseId.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocMillis.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalMillis.name()));
            assertEquals("[1]", customerMap.get(QueryField.shape.name()));
            assertEquals("1", customerMap.get(QueryField.eventCount.name()));
            assertEquals("" + pulse, customerMap.get(QueryField.value.name()));

            pulse = ++pulse % resultsPerChannel;
            if (pulse == 0) {
               ++channelCount;
               if (channelCount <= TEST_CHANNEL_NAMES.length + 1) {
                  // read comment (empty rows are skiped
                  mapReader.read("");
                  header = mapReader.getHeader(false);
               }
            }
         }
      } finally {
         mapReader.close();
      }
   }

   @Test
   public void testPulseRangeQueryWaveform() throws Exception {
      String channelName = "XYWaveform";
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            channelName);
      request.setResponseFormat(ResponseFormat.CSV);

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      LinkedHashSet<CellProcessor> cellProcessors = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.pulseId);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.shape);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.eventCount);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.value);
      cellProcessors.add(new NotNull());
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      CellProcessor[] processors = cellProcessors.toArray(new CellProcessor[cellProcessors.size()]);

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      ICsvMapReader mapReader =
            new CsvMapReader(new StringReader(response), CsvPreference.EXCEL_NORTH_EUROPE_PREFERENCE);
      try {
         // read comment
         mapReader.read("");
         final String[] header = mapReader.getHeader(false);
         Map<String, Object> customerMap;
         long pulse = 0;
         while ((customerMap = mapReader.read(header, processors)) != null) {
            assertEquals(channelName, customerMap.get(QueryField.channel.name()));
            assertEquals("" + pulse, customerMap.get(QueryField.pulseId.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocMillis.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalMillis.name()));
            assertEquals("[2048]", customerMap.get(QueryField.shape.name()));
            assertEquals("1", customerMap.get(QueryField.eventCount.name()));
            assertTrue(customerMap.get(QueryField.value.name()).toString().startsWith("["));
            assertTrue(customerMap.get(QueryField.value.name()).toString().endsWith("]"));

            ++pulse;
         }
      } finally {
         mapReader.close();
      }
   }

   @Test
   public void testTimeRangeQuery() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(0, 0),
                  TimeUtils.getTimeFromMillis(10, 0)),
            TEST_CHANNEL_NAMES);
      request.setResponseFormat(ResponseFormat.CSV);

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      LinkedHashSet<CellProcessor> cellProcessors = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.pulseId);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.shape);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.eventCount);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.value);
      cellProcessors.add(new NotNull());
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      CellProcessor[] processors = cellProcessors.toArray(new CellProcessor[cellProcessors.size()]);

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      CsvMapReader mapReader =
            new CsvMapReader(new StringReader(response), CsvPreference.EXCEL_NORTH_EUROPE_PREFERENCE);
      try {
         // read comment
         mapReader.read("");
         String[] header = mapReader.getHeader(false);
         Map<String, Object> customerMap;
         long resultsPerChannel = 2;
         long pulse = 0;
         long channelCount = 1;
         while ((customerMap = mapReader.read(header, processors)) != null) {
            assertEquals(TEST_CHANNEL + channelCount, customerMap.get(QueryField.channel.name()));
            assertEquals("" + pulse, customerMap.get(QueryField.pulseId.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocMillis.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalMillis.name()));
            assertEquals("[1]", customerMap.get(QueryField.shape.name()));
            assertEquals("1", customerMap.get(QueryField.eventCount.name()));
            assertEquals("" + pulse, customerMap.get(QueryField.value.name()));

            pulse = ++pulse % resultsPerChannel;
            if (pulse == 0) {
               ++channelCount;
               if (channelCount <= TEST_CHANNEL_NAMES.length) {
                  // read comment (empty rows are skiped
                  mapReader.read("");
                  header = mapReader.getHeader(false);
               }
            }
         }
      } finally {
         mapReader.close();
      }
   }

   @Test
   public void testDateRangeQuery() throws Exception {
      String startDate = RequestRangeDate.format(0);
      String endDate = RequestRangeDate.format(10);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            TEST_CHANNEL_NAMES);
      request.setResponseFormat(ResponseFormat.CSV);

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      LinkedHashSet<CellProcessor> cellProcessors = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.pulseId);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.shape);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.eventCount);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.value);
      cellProcessors.add(new NotNull());
      request.setFields(queryFields);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      CellProcessor[] processors = cellProcessors.toArray(new CellProcessor[cellProcessors.size()]);

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      ICsvMapReader mapReader =
            new CsvMapReader(new StringReader(response), CsvPreference.EXCEL_NORTH_EUROPE_PREFERENCE);
      try {
         // read comment
         mapReader.read("");
         String[] header = mapReader.getHeader(false);
         Map<String, Object> customerMap;
         long resultsPerChannel = 2;
         long pulse = 0;
         long channelCount = 1;
         while ((customerMap = mapReader.read(header, processors)) != null) {
            assertEquals(TEST_CHANNEL + channelCount, customerMap.get(QueryField.channel.name()));
            assertEquals("" + pulse, customerMap.get(QueryField.pulseId.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocMillis.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalMillis.name()));
            assertEquals("[1]", customerMap.get(QueryField.shape.name()));
            assertEquals("1", customerMap.get(QueryField.eventCount.name()));
            assertEquals("" + pulse, customerMap.get(QueryField.value.name()));

            pulse = ++pulse % resultsPerChannel;
            if (pulse == 0) {
               ++channelCount;
               if (channelCount <= TEST_CHANNEL_NAMES.length) {
                  // read comment (empty rows are skiped
                  mapReader.read("");
                  header = mapReader.getHeader(false);
               }
            }
         }
      } finally {
         mapReader.close();
      }
   }

   @Test
   public void testExtremaAggregation() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            Ordering.asc,
            AggregationType.extrema,
            TEST_CHANNEL_NAMES[0]);
      request.setResponseFormat(ResponseFormat.CSV);

      String content = mapper.writeValueAsString(request);

      try {
         this.mockMvc
               .perform(MockMvcRequestBuilders
                     .post(QueryRestController.QUERY)
                     .contentType(MediaType.APPLICATION_JSON)
                     .content(content))
               .andDo(MockMvcResultHandlers.print());
         // .andExpect(MockMvcResultMatchers.model().attribute(
         // "exception",
         // Matchers.isA(IllegalArgumentException.class)));
         // should throw an exception
         assertTrue(false);
      } catch (Throwable e) {
         assertTrue(true);
      }
   }

   @Test
   public void testIndexAggregation() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  0,
                  1),
            Ordering.asc,
            AggregationType.index,
            TEST_CHANNEL_NAMES[0]);
      request.setResponseFormat(ResponseFormat.CSV);

      String content = mapper.writeValueAsString(request);

      try {
         this.mockMvc
               .perform(MockMvcRequestBuilders
                     .post(QueryRestController.QUERY)
                     .contentType(MediaType.APPLICATION_JSON)
                     .content(content))
               .andDo(MockMvcResultHandlers.print());
         // .andExpect(MockMvcResultMatchers.model().attribute(
         // "exception",
         // Matchers.isA(IllegalArgumentException.class)));
         // should throw an exception
         assertTrue(false);
      } catch (Throwable e) {
         assertTrue(true);
      }
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregate() throws Exception {
      long startTime = 0;
      long endTime = 99;
      String startDate = RequestRangeDate.format(startTime);
      String endDate = RequestRangeDate.format(endTime);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            TEST_CHANNEL_01);
      request.setNrOfBins(2);
      request.setResponseFormat(ResponseFormat.CSV);

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      LinkedHashSet<CellProcessor> cellProcessors = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.pulseId);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.shape);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.eventCount);
      cellProcessors.add(new NotNull());
      request.setFields(queryFields);

      List<Aggregation> aggregations = new ArrayList<>();
      aggregations.add(Aggregation.min);
      cellProcessors.add(new NotNull());
      aggregations.add(Aggregation.mean);
      cellProcessors.add(new NotNull());
      aggregations.add(Aggregation.max);
      cellProcessors.add(new NotNull());
      request.setAggregations(aggregations);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      CellProcessor[] processors = cellProcessors.toArray(new CellProcessor[cellProcessors.size()]);

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      ICsvMapReader mapReader =
            new CsvMapReader(new StringReader(response), CsvPreference.EXCEL_NORTH_EUROPE_PREFERENCE);
      try {
         // read comment
         mapReader.read("");
         String[] header = mapReader.getHeader(false);
         Map<String, Object> customerMap;
         long pulse = 0;
         while ((customerMap = mapReader.read(header, processors)) != null) {
            assertEquals(TEST_CHANNEL_01, customerMap.get(QueryField.channel.name()));
            assertEquals("" + pulse, customerMap.get(QueryField.pulseId.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocMillis.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalMillis.name()));
            assertEquals("[1]", customerMap.get(QueryField.shape.name()));
            assertEquals("5", customerMap.get(QueryField.eventCount.name()));
            assertEquals("" + pulse + ".0", customerMap.get(Aggregation.min.name()));
            assertEquals("" + (pulse + 2) + ".0", customerMap.get(Aggregation.mean.name()));
            assertEquals("" + (pulse + 4) + ".0", customerMap.get(Aggregation.max.name()));

            pulse += 5;
         }
      } finally {
         mapReader.close();
      }
   }

   @Test
   public void testDateRangeQueryBinSizeAggregate() throws Exception {
      long startTime = 0;
      long endTime = 999;
      String startDate = RequestRangeDate.format(startTime);
      String endDate = RequestRangeDate.format(endTime);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            TEST_CHANNEL_01);
      request.setBinSize(100);
      request.setResponseFormat(ResponseFormat.CSV);

      LinkedHashSet<QueryField> queryFields = new LinkedHashSet<>();
      LinkedHashSet<CellProcessor> cellProcessors = new LinkedHashSet<>();
      queryFields.add(QueryField.channel);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.pulseId);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.iocMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalSeconds);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.globalMillis);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.shape);
      cellProcessors.add(new NotNull());
      queryFields.add(QueryField.eventCount);
      cellProcessors.add(new NotNull());
      request.setFields(queryFields);

      List<Aggregation> aggregations = new ArrayList<>();
      aggregations.add(Aggregation.min);
      cellProcessors.add(new NotNull());
      aggregations.add(Aggregation.mean);
      cellProcessors.add(new NotNull());
      aggregations.add(Aggregation.max);
      cellProcessors.add(new NotNull());
      request.setAggregations(aggregations);


      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      MvcResult result = this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      CellProcessor[] processors = cellProcessors.toArray(new CellProcessor[cellProcessors.size()]);

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      ICsvMapReader mapReader =
            new CsvMapReader(new StringReader(response), CsvPreference.EXCEL_NORTH_EUROPE_PREFERENCE);
      try {
         // read comment
         mapReader.read("");
         String[] header = mapReader.getHeader(false);
         Map<String, Object> customerMap;
         long pulse = 0;
         while ((customerMap = mapReader.read(header, processors)) != null) {
            assertEquals(TEST_CHANNEL_01, customerMap.get(QueryField.channel.name()));
            assertEquals("" + pulse, customerMap.get(QueryField.pulseId.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.iocMillis.name()));
            assertEquals("" + TimeUtils.getTimeStr(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalSeconds.name()));
            assertEquals("" + TimeUtils.getMillis(TestTimeUtils.getTimeFromPulseId(pulse)),
                  customerMap.get(QueryField.globalMillis.name()));
            assertEquals("[1]", customerMap.get(QueryField.shape.name()));
            assertEquals("10", customerMap.get(QueryField.eventCount.name()));
            assertEquals("" + pulse + ".0", customerMap.get(Aggregation.min.name()));
            assertEquals("" + (pulse + 4.5), customerMap.get(Aggregation.mean.name()));
            assertEquals("" + (pulse + 9) + ".0", customerMap.get(Aggregation.max.name()));

            pulse += 10;
         }
      } finally {
         mapReader.close();
      }
   }

   @Test
   public void testGzipFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setResponseFormat(ResponseFormat.CSV);
      request.setCompression(Compression.GZIP);

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.gz"));
   }

   @Test
   public void testCvsFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setResponseFormat(ResponseFormat.CSV);

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.csv"));
   }
}
